module vga11 (clk,r,g,b,hs,vs,vgaclk,       // vga interface
             addr,datain,dataout,wr,rd,ready,      // cpu interface
             sram_addr,sram_ce_n,sram_dq,sram_lb_n,sram_oe_n,sram_ub_n,sram_we_n,
				 debug,hbl,vbl,dli);// sram interface); 
				 // 1440x900 160x50 char with border text mode vga driver for DE2-115

// ver 0.20 alpha --- 2014.10.06
// (c) Piotr Kardasz pik33@we.pb.edu.pl
// This is free software, license: LGPL v.2.0 or newer

output reg [7:0] r,g,b;
output reg hs,vs;
output wire vgaclk; 
input wire clk;  //input clock=2x vgaclk
output wire ready;

output reg[7:0] debug;

reg [9:0] e;
reg f,oldwr,oldwr2;
reg [7:0] brr,brb,brg;
reg [19:0] caddr;
reg [31:0] lcnt;
reg [0:2] lstate,pstate; // 00 vis, 01 fp, 10 sync, 11 bp 100 border
reg [11:0] pcnt;
reg [31:0] fcnt; //frame counter

input wire [20:0] addr;
input wire [31:0] datain;
output reg [31:0] dataout;
input wire wr,rd;
output reg [19:0] sram_addr;
inout wire [15:0] sram_dq;
output wire sram_ce_n;
output reg sram_lb_n;
output wire sram_oe_n;
output reg sram_ub_n;
output reg sram_we_n;
output reg hbl,vbl,dli;
reg [15:0] sramdqout,dataout2,dataout3;
reg[31:0] datain2,datain3;
reg [20:0] addr2,sram_addr2;
reg[3:0] pcnt_mod,pcnt_mod2;
reg[7:0] pcnt_char;
reg[15:0] pixels3, pixels2;
reg[31:0] pixels,bb,cc,ss,mode,linebase,dl;
reg wr2,wr3,rd2,rd3,oldrd;
reg[3:0]wstate;
reg add;
(* ramstyle = "M9K" *) reg [31:0] pallette[1024]; // pallette registers
(* ramstyle = "M9K" *) reg [7:0] fontdef[4096];  // character generator
(* ramstyle = "M9K" *) reg [31:0] sprite[4096];  // sprite generator

reg[31:0] regs[24];  // internal regs

// 0 - display state: {vsync,hsync,frame counter}
// 1 - display start address
// 2 - mode
// 3 - border
// 4 - pallette bank
// 5 - mode 0 - 4 sprites @ full color; 1 - 16 sprites @ 256 colors; bits 3..2 - pallette selector
// 6 - display list start address; bit 31: dl active
// 7 - horizontal scroll
// 8..23 sprite x, sprite y, 2 high bits - width/height zoom

// display list 
// [31] - entry valid
// [30] - generate dli
// [29..22] palette index
// [21..18] mode
// [17..0] line start address; 




initial
  begin
  $readmemh("st4font.hex",fontdef);
  end

  
assign vgaclk=clk;
assign sram_ce_n=0;
assign sram_oe_n=0;
 

// ------------------------------------------- 
// bidir assignment
// ------------------------------------------- 

assign sram_dq = (sram_we_n==0) ? sramdqout : 16'bz;

//---------------------------------------------  
// debug  
//---------------------------------------------  
  
always@(posedge clk) 
 
begin 

debug[3:0]<=pcnt_mod;
debug[7]<=dl[31];
debug[6:4]<=0;
end



//---------------------------------------------
// main loop
//--------------------------------------------- 
 
always@(posedge clk)

  begin
  oldwr<=wr;
  if ((oldwr==0) && (wr==1)) wr2<=1; else wr2<=0;
  wr3<=wr2;
  oldrd<=rd;
  if ((oldrd==0) && (rd==1)) rd2<=1; else rd2<=0;
  rd3<=rd2;

  pcnt<=pcnt+1;
  pcnt_mod<=(regs[7]+pcnt+1)&15;
  pcnt_char<=(regs[7]+pcnt-143)>>4;  
  pcnt_mod2<=pcnt_mod-1;  
  
  
// ------------------------------------------- 
// sram rw state machine
// -------------------------------------------
  
// new state calculation
  case (wstate) 
  
    0: begin
	   addr2<=addr;
		if (((wr2==1) || (wr3==1)) && (rd2==0) && (rd3==0) && (addr[19]==0)) wstate<=1;
		else if (((wr2==1) || (wr3==1)) && (rd2==0) && (rd3==0) && (addr[19]==1)) wstate<=3;
		else if (((rd2==1) || (rd3==1)) && (wr2==0) && (wr3==0) && (addr[19]==0)) wstate<=4;
		else if (((rd2==1) || (rd3==1)) && (wr2==0) && (wr3==0) && (addr[19]==1)) wstate<=5;
		else wstate<=0;
		datain2=datain;
		end
    1: begin  
		if (pcnt_mod2==0) wstate<=6; else wstate<=1; 
//		datain2=datain;
		sramdqout<=datain2[31:16];
		sram_addr2<={addr2,1'b0};
      end 
    6: begin		
	   case (pcnt_mod)
		1,2,3:
		    begin
	       sram_addr<=sram_addr2;
	       sram_we_n<=0;    
          sram_lb_n<=0;
          sram_ub_n<=0;  
 	       end		
		4: 
		    begin
			 sram_addr<=sram_addr2;
  	       sram_we_n<=0;    
          sram_lb_n<=1;
          sram_ub_n<=1;  
	       end 	
		5,6,7:   	   
		    begin
		    sram_addr<=sram_addr2+1;
  	       sram_we_n<=0;    
          sram_lb_n<=0;
          sram_ub_n<=0;  
   	    sramdqout<=datain2[15:0];
			 end
   	8:
		    begin		    
			 sram_addr<=sram_addr2+1;
			 sram_we_n<=1;    
          sram_lb_n<=1;
          sram_ub_n<=1;  
			 wstate<=0;
          end
		default:
			wstate<=0;
		endcase	
     end
  
	4:	begin
    	if (pcnt_mod2==0) wstate<=10; else wstate<=4;
		sram_addr2<={addr2,1'b0};
      end		
	10:begin	
	   case (pcnt_mod2)
      1,2,3:
		    begin
	       sram_addr<=sram_addr2;
	       sram_we_n<=1;    
          sram_lb_n<=0;
          sram_ub_n<=0;  
 	       end
		4:
         begin
   	   sram_addr<=sram_addr2+1;
	      dataout3<=sram_dq;
  	      sram_we_n<=1;    
         sram_lb_n<=0;
         sram_ub_n<=0;  
	      end 	 

	   5,6,7:begin
   	   sram_addr<=sram_addr2+1;
  	      sram_we_n<=1;    
         sram_lb_n<=0;
         sram_ub_n<=0;  
	      end		
		
   	8:
         begin
	      dataout2<=sram_dq;
	      end 	 			
		9:
		   begin
		   dataout<={dataout3,dataout2};	 
	      wstate<=0;
			end
	   default:
         wstate<=0;
	      		
			
		endcase	
      end
		
   3: begin  
      wstate<=0;
      if (addr[15:12]==4'b0000) pallette[addr2[9:0]]<=datain; 
		else if (addr[15:12]==4'b0001)fontdef[addr2[10:0]]<=datain[7:0];
		else if (addr[15:12]==4'b0010)sprite[addr2[10:0]]<=datain;
		else regs [addr2[10:0]]<=datain;
   	end
		
	5: begin	

  	   wstate<=0;
	   if (addr[15:12]==4'b0000) dataout<=pallette[addr2[9:0]]; 
		else if (addr[15:12]==4'b0001) dataout<={24'b0,fontdef[addr2[10:0]]};
		else if (addr[15:12]==4'b0010) dataout<=sprite[addr2[10:0]];
		else dataout<=regs [addr[10:0]];
	   end
	 
   default: wstate<=0;
	
	endcase
	
//----- end of sram interface state machine

//----- vga 
	 
  if (pcnt_mod==9)	 	 
    begin
	 sram_addr<=caddr;
	 sram_we_n<=1;    
    sram_lb_n<=0;
    sram_ub_n<=0;  
	 end 
	 
  if (pcnt_mod==12)	 	 
    begin
	 sram_addr<=caddr+1;
	 pixels2<=sram_dq;
	 sram_we_n<=1;    
    sram_lb_n<=0;
    sram_ub_n<=0;  
	 end 

  if (pcnt_mod==15)	 	 
    begin
	 pixels3<=sram_dq;
	 end 	 
	 
  if (pcnt_mod==0) 
    begin
	 pixels={pixels2,pixels3};
	 if (pcnt>2016) dl<={pixels2,pixels3};
    end

  case (pcnt)

	 2048:if ((regs[6][31]==1)&&(dl[31]==1)) dli<=1;
	 
	 17:dli<=0;
    19:if ((regs[6][31]==1)&&(dl[31]==1)) regs[4]<={24'b0,dl[29:22]}; else regs[4]<=regs[4];
	 20:if ((regs[6][31]==1)&&(dl[31]==1)) regs[2]<={28'b0,dl[21:18]}; else regs[2]<=regs[2];
	 21:dl<=0;
    34:bb<=regs[3];
    35:mode<=regs[2];


  endcase 
  
  
  
  

 
 
// sprite machine 
 
    // line and pixel counters & state determination
  if (pcnt==96) begin pstate<=4; hbl<=0; end  // left border
  if (pcnt==159) pstate<=0; // display
  if (pcnt==1952) begin pstate<=4; end;// right border
  if (pcnt==2016) begin pstate<=1; hbl<=1; end // front porch
  if (pcnt==2064) begin pstate<=2; hs<=0; end// hsync
  if (pcnt==2096)          // back porch 
	 begin 
	 pstate<=3; 
    hs<=1;
	 pcnt<=1;
	 lcnt<=lcnt+1; 
    regs[0]<=lcnt+1;

    if ((dl[31]==1) && (regs[6][31]==1)) linebase<=dl[17:0]<<2; 
    else if (lcnt==65) linebase<=regs[1]<<1;
	 else if (mode==0)  linebase<=linebase+224;
	 else if ((mode==1) && (lcnt[0]==1'b1)) linebase<=linebase+224;
	 else if ((mode==2) && (lcnt[1:0]==2'b01)) linebase<=linebase+224;

	 else linebase<=linebase;
		
	 if (lcnt==25) begin vbl<=0; lstate<=4; end   // top border
	 else if (lcnt==65) begin  lstate<=0; linebase<=(regs[1]<<1); end   // display
	 else if (lcnt==1185) lstate<=4;  // bottom border
	 else if (lcnt==1225) begin lstate<=1; vbl<=1;  end  // front porch v
	 else if (lcnt==1228) begin lstate<=2; vs<=1; end // vsync 
	 else if (lcnt==1234) begin lstate<=3; vs<=0; lcnt<=0; fcnt<=fcnt+1;  end // back porch v
	 else begin lstate<=lstate; vs<=vs; end
	 
	 end
 
//    caddr=((lcnt-66)>>mode)*224+(pcnt_char<<1)+(regs[1]<<1);
    if (pcnt<2048) caddr=linebase+(pcnt_char<<1); else caddr<=((regs[6][19:0]+lcnt)<<1);

	 if (pstate==1)
	   begin
		brr=bb[7:0];
		brg=bb[15:8];
		brb=bb[23:16];
      end 
    
    if (((lstate==4)&&((pstate&3)==0)) || ((pstate==4)&&(lstate==0))) // borders
      begin 
      r<=brr; g<=brg; b<=brb;	
      end		 
	 
    else if (lstate==0 && pstate==0) // active display
      begin


		if (mode==0)
        case (pcnt_mod)
          0:e={regs[4][7:0],pixels[31:30]};
	  	    1:e={regs[4][7:0],pixels[29:28]};
		    2:e={regs[4][7:0],pixels[27:26]};
		    3:e={regs[4][7:0],pixels[25:24]};
		    4:e={regs[4][7:0],pixels[23:22]};
		    5:e={regs[4][7:0],pixels[21:20]};
		    6:e={regs[4][7:0],pixels[19:18]};
		    7:e={regs[4][7:0],pixels[17:16]};		            
			 8:e={regs[4][7:0],pixels[15:14]};
	  	    9:e={regs[4][7:0],pixels[13:12]};
		    10:e={regs[4][7:0],pixels[11:10]};
		    11:e={regs[4][7:0],pixels[9:8]};
		    12:e={regs[4][7:0],pixels[7:6]};
		    13:e={regs[4][7:0],pixels[5:4]};
		    14:e={regs[4][7:0],pixels[3:2]};
		    15:e={regs[4][7:0],pixels[1:0]};		  
		  endcase
		else if (mode==1)  
		   case (pcnt_mod)
  		   0,1:e={regs[4][5:0],pixels[31:28]};
		   2,3:e={regs[4][5:0],pixels[27:24]};
		   4,5:e={regs[4][5:0],pixels[23:20]};
		   6,7:e={regs[4][5:0],pixels[19:16]};  		   
			8,9:e={regs[4][5:0],pixels[15:12]};
		   10,11:e={regs[4][5:0],pixels[11:8]};
		   12,13:e={regs[4][5:0],pixels[7:4]};
		   14,15:e={regs[4][5:0],pixels[3:0]};
       endcase
	
   	else   
		   case (pcnt_mod)
  		   0,1,2,3:e={regs[4][1:0],pixels[31:24]};
		   4,5,6,7:e={regs[4][1:0],pixels[23:16]};
		   8,9,10,11:e={regs[4][1:0],pixels[15:8]};
		   12,13,14,15:e={regs[4][1:0],pixels[7:0]};  		   
        endcase


		cc<=pallette[e]; 
      if (pcnt>160) begin r=cc[7:0]; g=cc[15:8]; b=cc[23:16]; end
		else begin r=brr; g=brg; b=brb; end
		end
	 
	 else begin r<=0; g<=0; b<=0; end		

  end
 
endmodule 